// 4Balls.c

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#include "slib.h"
#include "wave.h"

#pragma resource "4balls.res"

#define ACCEL_LIMIT   15

#define COUNTER_ITERS 30

#define COUNTER_LIMIT 4000000

#define TIMER_TIMEOUT 70

#define MAXBALLS 4
                  
typedef struct _BALLINFO 
{
    int width;           // ball image width
    int height;          // ball image height
    int x;               // current x position of image's upper-left corner
    int y;               // current y position of image's upper-left corner

    int dx;              // ball's current horizontal velocity
    int dy;              // ball's current vertical velocity

    HBITMAP g_hbmBall;   // ball image bitmap handle

    int acceleration;    // ball's change in velocity
}
BALLINFO;

// globals

BALLINFO g_ballInfo [MAXBALLS]; // ball information structures

static BOOL g_bMuted = TRUE;    // sound muted flag (mute sound when TRUE)

static HBITMAP g_hbmMask;       // ball image mask to prevent transparency

static int g_iCounter;          // acceleration change counter 

static int g_iSndID;            // ball sound effect ID

// local function prototypes

void DrawBalls   (HDC, RECT *);
void ReadConfig  (void);
void UpdateBalls (RECT *);
void WriteConfig (void);
                                   
// =========================================================================
// BOOL WINAPI RegisterDialogClasses (HANDLE hinstance)
//
// Allow the programmer to register any child control windows.
//
// Arguments:
//
// hinstance - screensaver instance handle
//
// Return:
//
// TRUE if child control windows successfully created or no controls needed;
// otherwise FALSE
// =========================================================================

#pragma argsused
BOOL WINAPI RegisterDialogClasses (HANDLE hinstance)
{
   return TRUE; // no controls need to be registered.
}

// ===========================================================================
// BOOL WINAPI ScreenSaverConfigureDialog (HWND hwnd, int iMsg, WPARAM wparam,
//                                         LPARAM lparam)
//
// Screensaver-specific configuration procedure for processing messages.
//
// Arguments:
//
// hwnd   - handle of window associated with this screensaver procedure
//
// iMsg   - message identifier
//
// wparam - 32-bit word parameter with message data (if applicable)
//
// lparam - 32-bit long word parameter with message data (if applicable)
//
// Return:
//
// TRUE if message processed; otherwise FALSE if message not processed or
// Windows should not set the default keyboard focus (WM_INITDIALOG)
// ===========================================================================

#pragma argsused
BOOL WINAPI ScreenSaverConfigureDialog (HWND hwnd, UINT iMsg, WPARAM wparam,
                                        LPARAM lparam)
{
   switch (iMsg)
   {
      case WM_COMMAND:
           if (wparam == IDOK || wparam == IDCANCEL)
           {
               EndDialog (hwnd, 0);

               return TRUE;
           }
   }

   return FALSE;
}

// =====================================================================
// LRESULT WINAPI ScreenSaverProc (HWND hwnd, int iMsg, WPARAM wparam,
//                                 LPARAM lparam)
//
// Screensaver-specific procedure for processing messages.
//
// Arguments:
//
// hwnd   - handle of window associated with this screensaver procedure
//
// iMsg   - message identifier
//
// wparam - 32-bit word parameter with message data (if applicable)
//
// lparam - 32-bit long word parameter with message data (if applicable)
//
// Return:
//
// 0 if message processed, otherwise DefScreenSaverProc() result
// =====================================================================

LRESULT WINAPI ScreenSaverProc (HWND hwnd, UINT iMsg, WPARAM wparam,
                                LPARAM lparam)
{
   int i;
   BITMAP bm;
   char *szWaveName;
   static UINT uTimerID;

   switch (iMsg)
   {
      case WM_CREATE:
           // Create an animation timer that results in a WM_TIMER message
           // approximately every TIMER_TIMEOUT milliseconds.

           if ((uTimerID = SetTimer (hwnd, 1, TIMER_TIMEOUT, NULL)) == NULL)
           {
               MessageBox (hwnd, "No available timer!", "4balls", MB_OK);
               return -1;
           }

           // Attempt to open the ball sound effect. If successful, clear the
           // global sound mute flag to FALSE -- do not mute sound.

           if ((g_iSndID = OpenWave ("ballsound")) >= 0)
               g_bMuted = FALSE;

           // Load the ball image mask, which is a solid black ball image on a
           // white background.

           g_hbmMask = LoadBitmap (GetModuleHandle (NULL), "mask");

           // Ensure that program doesn't always start with the same random
           // number sequence.

           randomize ();

           // Initialize all ball information structures.

           for (i = 0; i < MAXBALLS; i++)
           {
                // Zero out each ball's information structure.

                ZeroMemory (&g_ballInfo [i], sizeof(g_ballInfo [i]));

                // Select the ball's bitmap resource name.

                switch (i)
                {
                   case 0 : szWaveName = "ballblu"; break;
                   case 1 : szWaveName = "ballgrn"; break;
                   case 2 : szWaveName = "ballmag"; break;
                   default: szWaveName = "ballora"; 
                }

                // Load the ball's bitmap resource. This resource defines the
                // ball's image on a black background.

                g_ballInfo [i].g_hbmBall = LoadBitmap (GetModuleHandle (NULL),
                                                       szWaveName);

                // Obtain ball image's width and height.

                GetObject (g_ballInfo [i].g_hbmBall, sizeof(bm), &bm);

                // Save width and height for use in DrawBalls() and
                // UpdateBalls().

                g_ballInfo [i].width = bm.bmWidth;
                g_ballInfo [i].height = bm.bmHeight;

                // Initialize ball's acceleration and velocity.

                g_ballInfo [i].acceleration = i*MAXBALLS+rand ()%MAXBALLS+1;
                g_ballInfo [i].dx = g_ballInfo [i].acceleration;
                g_ballInfo [i].dy = g_ballInfo [i].acceleration;
           }

           // Read configuration data from Windows registry.

           ReadConfig ();
           return 0;

      case WM_DESTROY:
           // Close the ball sound effect.

           if (g_iSndID >= 0)
               CloseWave (g_iSndID);

           // Destroy the timer.

           KillTimer (hwnd, uTimerID);

           // Inform Windows that screensaver is terminating.

           PostQuitMessage (0);
           return 0;

      case WM_TIMER:
      {
           RECT rcClient;
           HDC hdc = GetDC (hwnd);

           // Get screensaver window dimensions.

           GetClientRect (hwnd, &rcClient);

           // Update ball positions.

           UpdateBalls (&rcClient);

           // Draw balls in the screensaver window.

           DrawBalls (hdc, &rcClient);

           ReleaseDC (hwnd, hdc);

           // Update the acceleration counter. Prevent negative numbers by
           // resetting the counter when it reaches a positive limit. You can
           // choose any positive number you want by modifying COUNTER_LIMIT.

           g_iCounter++;
           if (g_iCounter > COUNTER_LIMIT)
               g_iCounter = 0;

           return 0;
      }
   }

   return DefScreenSaverProc (hwnd, iMsg, wparam, lparam);
}

// ===================================================
// void DrawBalls (HDC hdc, RECT *prc)
//
// Draw balls in the screensaver window.
//
// Arguments:
//
// hdc - screensaver window device context
// prc - pointer to screenwindow client area rectangle
//
// Return:
//
// none
// ===================================================

void DrawBalls (HDC hdc, RECT *prc)
{
   // Setup.

   HDC hdcBuffer = CreateCompatibleDC (hdc);

   HBITMAP hbmBuffer = CreateCompatibleBitmap (hdc, prc->right, prc->bottom);

   HBITMAP hbmOldBuffer = SelectObject (hdcBuffer, hbmBuffer);

   HDC hdcMem = CreateCompatibleDC (hdc);

   HBITMAP hbmOld = NULL;

   // Paint background buffer using a blue/gold vertical gradient.

   int r1 = 5, g1 = 35, b1 = 145;
   int r2 = 167, g2 = 167, b2 = 76;

   int i;
   for (i = 0; i < prc->bottom-prc->top+1; i++)
   {
        HBRUSH hbr;
        int r, g, b;
        RECT rectTemp;

        r = r1+(i*(r2-r1)/(prc->bottom-prc->top+1));
        g = g1+(i*(g2-g1)/(prc->bottom-prc->top+1));
        b = b1+(i*(b2-b1)/(prc->bottom-prc->top+1));

        rectTemp.left = prc->left;
        rectTemp.right = prc->right+1;
        rectTemp.top = prc->top+i;
        rectTemp.bottom = prc->top+i+1;

        hbr = CreateSolidBrush (RGB (r, g, b));

        FillRect (hdcBuffer, &rectTemp, hbr);

        DeleteObject (hbr);
   }                

   // Draw the balls.

   for (i = 0; i < MAXBALLS; i++)
   {
        // Preserve window background surrounding the ball image. SRCAND ands
        // the mask's white background, whose bits are ones, with whatever
        // bits constitute the window background surrounding the ball image.
        // This preserves that background.

        // Remove background under the ball image. SRCAND ands the mask's
        // black ball image, whose bits are zeroes, with whatever bits
        // constitute the background under the window's ball image. This
        // results in a black round hole in the window.

        // The SRCAND operation is necessary to prevent a transparency effect.
        // If this operation was absent, the ball images would appear in the
        // window, but the window's background would show through the ball
        // images. Furthermore, if one of the ball images passed over another
        // ball image, the underlying ball image would show through. 

        SelectObject (hdcMem, g_hbmMask);
        BitBlt (hdcBuffer, g_ballInfo [i].x, g_ballInfo [i].y,
                g_ballInfo [i].width, g_ballInfo [i].height, hdcMem, 0, 0,
                SRCAND);

        // Preserve window background surrounding the ball image. SRCPAINT ors 
        // the ball image's black background, whose bits are zeroes, with
        // whatever bits constitute the window background surrounding the ball
        // image. This preserves that background.

        // Merge ball image with window's black round hole. SRCPAINT ors the
        // ball image's bits with the zero bits constituting the black round
        // hole, resulting in a solid ball image with no transparency effect.

        SelectObject (hdcMem, g_ballInfo [i].g_hbmBall);
        BitBlt (hdcBuffer, g_ballInfo [i].x, g_ballInfo [i].y,
                g_ballInfo [i].width, g_ballInfo [i].height, hdcMem, 0, 0,
                SRCPAINT);

        SelectObject (hdcMem, g_hbmMask);
   }

   // Update screensaver window with background buffer.

   BitBlt (hdc, 0, 0, prc->right, prc->bottom, hdcBuffer, 0, 0, SRCCOPY);

   // Cleanup.

   SelectObject (hdcMem, hbmOld);
   DeleteDC (hdcMem);

   SelectObject (hdcBuffer, hbmOldBuffer);

   DeleteObject (hbmBuffer);

   DeleteDC (hdcBuffer);
}

// ====================================================
// void ReadConfig (void)
//
// Read configuation information from Windows registry.
//
// Arguments:
//
// none
//
// Return:
//
// none
// ====================================================

static void ReadConfig (void)
{
}

// ===================================================
// void UpdateBalls (RECT *prc)
//
// Update ball positions.
//
// Arguments:
//
// prc - pointer to screenwindow client area rectangle
//
// Return:
//
// none
// ===================================================

void UpdateBalls (RECT *prc)
{
   int i;
   for (i = 0; i < MAXBALLS; i++)
   {
        // Modify the acceleration every COUNTER_ITERS iterations. When the
        // ball's acceleration exceeds ACCEL_LIMIT, reset the acceleration to
        // 1 so that balls don't constantly accelerate.

        if (g_iCounter % COUNTER_ITERS == 0)
        {
            g_ballInfo [i].acceleration++;
            if (g_ballInfo [i].acceleration > ACCEL_LIMIT)
                g_ballInfo [i].acceleration = 1;
        }

        // Move the ball to a new position based on its velocity.

        g_ballInfo [i].x += g_ballInfo [i].dx;
        g_ballInfo [i].y += g_ballInfo [i].dy;

        // When the ball reaches the left or right edges of the screensaver
        // window, reverse the ball's direction, accelerate/decelerate the
        // ball, and play a sound (if sound isn't muted) to indicate a bounce.

        if (g_ballInfo [i].x < 0)
        {
            g_ballInfo [i].x = 0;
            g_ballInfo [i].dx = g_ballInfo [i].acceleration;

            if (!g_bMuted)
                PlayWave (g_iSndID);
        }
        else
        if (g_ballInfo [i].x + g_ballInfo [i].width > prc->right)
        {
            g_ballInfo [i].x = prc->right - g_ballInfo [i].width;
            g_ballInfo [i].dx = -g_ballInfo [i].acceleration;

            if (!g_bMuted)
                PlayWave (g_iSndID);
        }

        // When the ball reaches the top or bottom edges of the screensaver
        // window, reverse the ball's direction, accelerate/decelerate the
        // ball, and play a sound (if sound isn't muted) to indicate a bounce.

        if (g_ballInfo [i].y < 0)
        {
            g_ballInfo [i].y = 0;
            g_ballInfo [i].dy = g_ballInfo [i].acceleration;

            if (!g_bMuted)
                PlayWave (g_iSndID);
        }
        else
        if (g_ballInfo [i].y + g_ballInfo [i].height > prc->bottom)
        {
            g_ballInfo [i].y = prc->bottom - g_ballInfo [i].height;
            g_ballInfo [i].dy = -g_ballInfo [i].acceleration;

            if (!g_bMuted)
                PlayWave (g_iSndID);
        }
   }
}

// ===================================================
// void WriteConfig (void)
//
// Write configuation information to Windows registry.
//
// Arguments:
//
// none
//
// Return:
//
// none
// ===================================================

static void WriteConfig (void)
{
}
